# -*- coding: utf-8 -*-
"""Module for storing type ids of common mincePy types"""
import uuid

FILE_TYPE_ID = uuid.UUID('3bf3c24e-f6c8-4f70-956f-bdecd7aed091')
OBJ_REF_TYPE_ID = uuid.UUID('633c7035-64fe-4d87-a91e-3b7abd8a6a28')
SNAPSHOT_ID_TYPE_ID = uuid.UUID('05fe092b-07b3-4ffc-8cf2-cee27aa37e81')
